"""Jedi Language Server

Creates the language server constant and wraps "features" with it.

Official language server spec:
    https://microsoft.github.io/language-server-protocol/specification
"""

from typing import Dict  # pylint: disable=unused-import
from typing import List, Optional

from pygls.features import (
    COMPLETION,
    DEFINITION,
    DOCUMENT_SYMBOL,
    HOVER,
    REFERENCES,
    RENAME,
    WORKSPACE_SYMBOL,
)
from pygls.server import LanguageServer
from pygls.types import (
    CompletionItem,
    CompletionList,
    CompletionParams,
    DocumentSymbolParams,
    Hover,
    Location,
    RenameParams,
    SymbolInformation,
    TextDocumentPositionParams,
    TextEdit,
    WorkspaceEdit,
    WorkspaceSymbolParams,
)

from .server_utils import (
    get_jedi_document_names,
    get_jedi_script,
    get_jedi_workspace_names,
    get_location_from_definition,
    get_symbol_information_from_definition,
)
from .type_map import get_lsp_completion_type

SERVER = LanguageServer()


@SERVER.feature(COMPLETION, triggerCharacters=["."])
def lsp_completion(server: LanguageServer, params: CompletionParams = None):
    """Returns completion items."""
    script = get_jedi_script(server, params)
    jedi_completions = script.completions()
    return CompletionList(
        is_incomplete=False,
        items=[
            CompletionItem(
                label=completion.name,
                kind=get_lsp_completion_type(completion.type),
                detail=completion.description,
                documentation=completion.docstring(),
                insert_text=completion.name,
            )
            for completion in jedi_completions
        ],
    )


@SERVER.feature(DEFINITION)
def lsp_definition(
    server: LanguageServer, params: TextDocumentPositionParams
) -> List[Location]:
    """Support Goto Definition"""
    script = get_jedi_script(server, params)
    definitions = script.goto_assignments(
        follow_imports=True, follow_builtin_imports=True
    )
    return [
        get_location_from_definition(definition) for definition in definitions
    ]


@SERVER.feature(HOVER)
def lsp_hover(
    server: LanguageServer, params: TextDocumentPositionParams
) -> Hover:
    """Support the hover feature"""
    script = get_jedi_script(server, params)
    definitions = script.goto_definitions()
    return Hover(
        contents=(
            definitions[0].docstring()
            if definitions
            else "No docstring definition found."
        )
    )


@SERVER.feature(REFERENCES)
def lsp_references(
    server: LanguageServer, params: TextDocumentPositionParams
) -> List[Location]:
    """Obtain all references to document"""
    script = get_jedi_script(server, params)
    try:
        definitions = script.usages()
    except Exception:  # pylint: disable=broad-except
        return []
    return [
        get_location_from_definition(definition) for definition in definitions
    ]


@SERVER.feature(RENAME)
def lsp_rename(
    server: LanguageServer, params: RenameParams
) -> Optional[WorkspaceEdit]:
    """Rename a symbol across a workspace"""
    script = get_jedi_script(server, params)
    try:
        definitions = script.usages()
    except Exception:  # pylint: disable=broad-except
        return None
    locations = [
        get_location_from_definition(definition) for definition in definitions
    ]
    if not locations:
        return None
    changes = {}  # type: Dict[str, List[TextEdit]]
    for location in locations:
        text_edit = TextEdit(location.range, new_text=params.newName)
        if location.uri not in changes:
            changes[location.uri] = [text_edit]
        else:
            changes[location.uri].append(text_edit)
    return WorkspaceEdit(changes=changes)


@SERVER.feature(DOCUMENT_SYMBOL)
def lsp_document_symbol(
    server: LanguageServer, params: DocumentSymbolParams
) -> List[SymbolInformation]:
    """Document Python document symbols"""
    jedi_names = get_jedi_document_names(server, params)
    return [
        get_symbol_information_from_definition(definition)
        for definition in jedi_names
    ]


@SERVER.feature(WORKSPACE_SYMBOL)
def lsp_workspace_symbol(
    server: LanguageServer,
    params: WorkspaceSymbolParams,  # pylint: disable=unused-argument
) -> List[SymbolInformation]:
    """Document Python workspace symbols"""
    jedi_names = get_jedi_workspace_names(server)
    return [
        get_symbol_information_from_definition(definition)
        for definition in jedi_names
    ]



"""Utility functions used by the language server"""

import itertools
import os
import subprocess
from typing import List, Optional, Union

import jedi.api
from jedi import Script
from jedi.api.classes import Definition
from jedi.api.environment import get_cached_default_environment
from pygls.server import LanguageServer, Workspace
from pygls.types import (
    DocumentSymbolParams,
    Location,
    Position,
    Range,
    RenameParams,
    SymbolInformation,
    TextDocumentItem,
    TextDocumentPositionParams,
)
from pygls.uris import from_fs_path

from .type_map import get_lsp_symbol_type


def get_jedi_script(
    server: LanguageServer,
    params: Union[TextDocumentPositionParams, RenameParams],
) -> Script:
    """Simplifies getting jedi Script

    NOTE:
        * jedi is 1-indexed for lines and 0-indexed for columns
        * LSP is 0-indexed for lines and 0-indexed for columns
        * Therefore, add 1 to LSP's request for the line
    """
    workspace = server.workspace
    text_doc = workspace.get_document(params.textDocument.uri)
    return Script(
        source=text_doc.source,
        path=text_doc.path,
        line=params.position.line + 1,
        column=params.position.character,
        environment=get_cached_default_environment(),
    )


def get_jedi_document_names(
    server: LanguageServer, params: DocumentSymbolParams,
) -> List[Definition]:
    """Get a list of document names from Jedi"""
    workspace = server.workspace
    text_doc = workspace.get_document(params.textDocument.uri)
    return jedi.api.names(
        source=text_doc.source,
        path=text_doc.path,
        all_scopes=True,
        definitions=True,
        references=False,
        environment=get_cached_default_environment(),
    )


class WorkspaceDocuments:
    """A class to manage one-time gathering of workspace documents"""

    # pylint: disable=too-few-public-methods

    def __init__(self) -> None:
        self.gathered_names = False

    def gather_workspace_names(self, workspace: Workspace) -> None:
        """Collect the workspace names"""
        if not self.gathered_names:
            git_path = os.path.join(workspace.root_path, ".git")
            git_command = "git --git-dir=" + git_path + " ls-files --full-name"
            try:
                git_output = subprocess.check_output(git_command, shell=True)
            except subprocess.CalledProcessError:
                self.gathered_names = True
                return
            project_paths = git_output.decode("utf-8").splitlines()
            for doc_path in project_paths:
                if os.path.splitext(doc_path)[1] == ".py":
                    full_doc_path = os.path.join(workspace.root_path, doc_path)
                    doc_uri = from_fs_path(full_doc_path)
                    with open(full_doc_path) as infile:
                        text = infile.read()
                    workspace.put_document(
                        TextDocumentItem(
                            uri=doc_uri,
                            language_id="python",
                            version=0,
                            text=text,
                        )
                    )
        self.gathered_names = True


_WORKSPACE_DOCUMENTS = WorkspaceDocuments()


def get_jedi_workspace_names(server: LanguageServer) -> List[Definition]:
    """Get a list of workspace names from Jedi"""
    workspace = server.workspace
    _WORKSPACE_DOCUMENTS.gather_workspace_names(workspace)
    definitions_by_document = (
        jedi.api.names(
            source=text_doc.source,
            path=text_doc.path,
            all_scopes=True,
            definitions=True,
            references=False,
            environment=get_cached_default_environment(),
        )
        for text_doc in workspace.documents.values()
        if text_doc and os.path.splitext(text_doc.path)[1] == ".py"
    )
    return list(
        itertools.chain.from_iterable(
            definitions if definitions else []
            for definitions in definitions_by_document
        )
    )


def get_location_from_definition(definition: Definition) -> Location:
    """Get LSP location from Jedi definition

    NOTE:
        * jedi is 1-indexed for lines and 0-indexed for columns
        * LSP is 0-indexed for lines and 0-indexed for columns
        * Therefore, subtract 1 from Jedi's definition line
    """
    return Location(
        uri=from_fs_path(definition.module_path),
        range=Range(
            start=Position(
                line=definition.line - 1, character=definition.column
            ),
            end=Position(
                line=definition.line - 1,
                character=definition.column + len(definition.name),
            ),
        ),
    )


def get_symbol_information_from_definition(
    definition: Definition,
) -> SymbolInformation:
    """Get LSP SymbolInformation from Jedi definition"""
    return SymbolInformation(
        name=definition.name,
        kind=get_lsp_symbol_type(definition.type),
        location=get_location_from_definition(definition),
        container_name=get_jedi_parent_name(definition),
    )


def get_jedi_parent_name(definition: Definition) -> Optional[str]:
    """Retrieve the parent name from Jedi

    Error prone Jedi calls are wrapped in try/except to avoid
    """
    try:
        parent = definition.parent()
    except Exception:  # pylint: disable=broad-except
        return None
    return parent.name if parent and parent.parent() else None
